/**
   Copyright 2010 Carlo Alberto Degli Atti (lord kada)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package it.concept.pin;

import it.concept.pin.annotations.Pin;
import it.concept.pin.annotations.Transport;
import it.concept.pin.exceptions.DataModelFieldDoesNotExist;
import it.concept.pin.exceptions.DataModelMethodDoesNotExist;
import it.concept.pin.exceptions.PinException;
import it.concept.pin.exceptions.SetterMethodIsNotVoid;
import it.concept.pin.exceptions.WrongNumberOfParametersInSetterMethod;
import it.concept.pin.proxy.ProxyWrapper;

import java.lang.annotation.AnnotationFormatError;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class PinManager {

	private static final String GETTER_METHOD_PREFIX = "get";
	
	private static final String SETTER_METHOD_PREFIX = "set";

	@SuppressWarnings("unchecked")
	public static Object toPojo( Class<?> pojoClass, Object model ) throws PinException {
		
		Object pojo = null;
		
		if ( model != null ) {
				
			try {
				
				pojo = pojoClass.newInstance();
				
				if ( pojoClass.getAnnotation( Pin.class ) == null )
					
					throw new AnnotationFormatError( pojoClass.getName() + " is not @Pin annotated" );
				
				// Let's process the class members annotated with @Transport
				for ( Field pojoField : getFields( pojoClass ) ) {
					
					pojoField.setAccessible( true );
					
					Transport pojoFieldAnnotation = pojoField.getAnnotation( Transport.class );
					
					if ( pojoFieldAnnotation != null ) {
						
						Field modelField = getField( pojoField.getName(), model.getClass() );

						if ( modelField != null ) {

							modelField.setAccessible( true );
							
							Object valueToSet = modelField.get( model );
							
							if ( valueToSet instanceof Collection<?> ) {
								
								if ( ! pojoFieldAnnotation.value().equals( Void.class ) ) {
									
									Collection<Object> newInstance = ( Collection<Object> ) valueToSet.getClass().newInstance();
									
									for ( Object item : ( Collection<Object> ) valueToSet )
										
										newInstance.add( toPojo( pojoFieldAnnotation.value(), item ) );
									
									valueToSet = newInstance;
									
								}
								
							}
							
							else
								
								if ( pojoField.getType().getAnnotation( Pin.class ) != null )
								
									valueToSet = toPojo( pojoField.getType(), valueToSet );
								
								else
							
									valueToSet = modelField.get( model );

							pojoField.set( pojo, valueToSet );

						}
						
						else 
							
							throw new DataModelFieldDoesNotExist( model.getClass().getName(), pojoField.getName() );
						
					}
					
				}
				
				// Let's process the setter methods that are annotated with @Transport
				// Note that a setter method: 
				// a) returns void b) expects one parameter c) name begins with 'set'
				for ( Method setterMethod : getMethods( pojoClass ) ) {
				
					Transport methodAnnotation = setterMethod.getAnnotation( Transport.class );
					
					if ( methodAnnotation != null ) {
						
						if ( setterMethod.getReturnType().equals( void.class ) ) {
						
							setterMethod.setAccessible( true );
						
							Class<?>[] parameterTypes = setterMethod.getParameterTypes();
						
							if ( parameterTypes.length != 1 )
								
								throw new WrongNumberOfParametersInSetterMethod( setterMethod.getName() );
							
							String modelGetterMethodName = setterMethod.getName().replaceAll( SETTER_METHOD_PREFIX, GETTER_METHOD_PREFIX );
							
							Method modelGetterMethod = getMethod( modelGetterMethodName, model.getClass() );
	
							if ( modelGetterMethod != null ) {
								
								Object valueToSet = modelGetterMethod.invoke( model, new Object[] { } );
								
								if ( valueToSet instanceof Collection<?> ) {
									
									if ( ! methodAnnotation.value().equals( Void.class ) ) {
										
										Collection<Object> newInstance = ( Collection<Object> ) valueToSet.getClass().newInstance();
										
										for ( Object item : ( Collection<Object> ) valueToSet )
											
											newInstance.add( toPojo( methodAnnotation.value(), item ) );
										
										valueToSet = newInstance;
										
									}
									
								}
								
								else 
									
									if ( parameterTypes[0].getAnnotation( Pin.class ) != null )

										valueToSet = toPojo( parameterTypes[0], valueToSet );
									
								setterMethod.invoke( pojo, new Object[] { valueToSet } );

							}
							
							else
								
								throw new DataModelMethodDoesNotExist( model.getClass().getName(), modelGetterMethodName );
								
						}
						
						else
							
							throw new SetterMethodIsNotVoid( setterMethod.getName() );
						
					}
					
				}
	
			}
			catch ( Exception e ) {
				
				throw new PinException( e );
				
			}
		
		}
			
		return pojo;
		
	}

	public static Object toInterface( Object pojo ) throws PinException {

		try {
			
			Class<?> pojoClass = pojo.getClass();
			
			Pin pinAnnotation = pojoClass.getAnnotation( Pin.class );
			
			if ( pinAnnotation == null )
				
				throw new AnnotationFormatError( pojoClass.getName() + " is not @Pin annotated" );
			
			return ProxyWrapper.createProxy( Class.forName( pinAnnotation.javaInterfaceName() ), pojo );
			
		} catch ( Exception e ) {
			
			throw new PinException( e );
			
		}
		
	}
	
	private static List<Field> getFields( Class<?> type ) {
		
		return _getFields( new ArrayList<Field>(), type );
		
	}
	
	private static List<Field> _getFields( List<Field> fields, Class<?> type ) {
		
	    for ( Field field: type.getDeclaredFields() )

	    	fields.add(field);

	    if ( type.getSuperclass() != null ) 
	    	
	        return _getFields( fields, type.getSuperclass() );

	    return fields;
	
	}
	
	private static Field getField( String fieldName, Class<?> type ) throws NoSuchFieldException {
		
		try {
			
		return type.getDeclaredField( fieldName );
		
		} catch( NoSuchFieldException e ) {

			Class<?> superclass = type.getSuperclass();
			
			if ( superclass != null) 
				
				return getField( fieldName, superclass );
			
			throw e;
			
		}
	
	}

	private static List<Method> getMethods( Class<?> type ) {
		
		return _getMethods( new ArrayList<Method>(), type );
		
	}
	
	private static List<Method> _getMethods( List<Method> methods, Class<?> type ) {
		
	    for ( Method method: type.getDeclaredMethods() )
	        
	    	if ( method.getName().startsWith( "set") )
	    			
	    			methods.add( method );

	    if ( type.getSuperclass() != null ) 
	    	
	        return _getMethods( methods, type.getSuperclass() );
	    
	    return methods;
	    
	}
	
	private static Method getMethod( String methodName, Class<?> type ) throws NoSuchMethodException {
		
		try {
			
		return type.getDeclaredMethod( methodName, new Class<?>[]{} );
		
		} catch( NoSuchMethodException e ) {

			Class<?> superclass = type.getSuperclass();
			
			if ( superclass != null) 
				
				return getMethod( methodName, superclass );
			
			throw e;
			
		}
	
	}
	
}